@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2013-2021 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
+ *  Copyright (C) 2013-2024 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
  *  Copyright (C) 2011-2013 Sourcefire, Inc.
  *
  *  Authors: aCaB <acab@clamav.net>
@@ -45,22 +45,24 @@ static const void *needblock(const iso9660_t *iso, unsigned int block, int temp)
     cli_ctx *ctx = iso->ctx;
     size_t loff;
     unsigned int blocks_per_sect = (2048 / iso->blocksz);
-    if (block > (((*ctx->fmap)->len - iso->base_offset) / iso->sectsz) * blocks_per_sect)
+    if (block > ((ctx->fmap->len - iso->base_offset) / iso->sectsz) * blocks_per_sect)
         return NULL;                                  /* Block is out of file */
     loff = (block / blocks_per_sect) * iso->sectsz;   /* logical sector */
     loff += (block % blocks_per_sect) * iso->blocksz; /* logical block within the sector */
     if (temp)
-        return fmap_need_off_once(*ctx->fmap, iso->base_offset + loff, iso->blocksz);
-    return fmap_need_off(*ctx->fmap, iso->base_offset + loff, iso->blocksz);
+        return fmap_need_off_once(ctx->fmap, iso->base_offset + loff, iso->blocksz);
+    return fmap_need_off(ctx->fmap, iso->base_offset + loff, iso->blocksz);
 }
 
-static int iso_scan_file(const iso9660_t *iso, unsigned int block, unsigned int len)
+static cl_error_t iso_scan_file(const iso9660_t *iso, unsigned int block, unsigned int len)
 {
     char *tmpf;
-    int fd, ret = CL_SUCCESS;
+    int fd         = -1;
+    cl_error_t ret = CL_SUCCESS;
 
-    if (cli_gentempfd(iso->ctx->sub_tmpdir, &tmpf, &fd) != CL_SUCCESS)
+    if (cli_gentempfd(iso->ctx->sub_tmpdir, &tmpf, &fd) != CL_SUCCESS) {
         return CL_ETMPFILE;
+    }
 
     cli_dbgmsg("iso_scan_file: dumping to %s\n", tmpf);
     while (len) {
@@ -81,8 +83,9 @@ static int iso_scan_file(const iso9660_t *iso, unsigned int block, unsigned int
         block++;
     }
 
-    if (!len)
-        ret = cli_magic_scan_desc(fd, tmpf, iso->ctx, iso->buf);
+    if (!len) {
+        ret = cli_magic_scan_desc(fd, tmpf, iso->ctx, iso->buf, LAYER_ATTRIBUTES_NONE);
+    }
 
     close(fd);
     if (!iso->ctx->engine->keeptmp) {
@@ -117,18 +120,17 @@ static char *iso_string(iso9660_t *iso, const void *src, unsigned int len)
     return iso->buf;
 }
 
-static int iso_parse_dir(iso9660_t *iso, unsigned int block, unsigned int len)
+static cl_error_t iso_parse_dir(iso9660_t *iso, unsigned int block, unsigned int len)
 {
-    cli_ctx *ctx      = iso->ctx;
-    int ret           = CL_CLEAN;
-    int viruses_found = 0;
+    cli_ctx *ctx   = iso->ctx;
+    cl_error_t ret = CL_SUCCESS;
 
     if (len < 34) {
         cli_dbgmsg("iso_parse_dir: Directory too small, skipping\n");
-        return CL_CLEAN;
+        return CL_SUCCESS;
     }
 
-    for (; len && ret == CL_CLEAN; block++, len -= MIN(len, iso->blocksz)) {
+    for (; len && ret == CL_SUCCESS; block++, len -= MIN(len, iso->blocksz)) {
         const uint8_t *dir, *dir_orig;
         unsigned int dirsz;
 
@@ -137,15 +139,18 @@ static int iso_parse_dir(iso9660_t *iso, unsigned int block, unsigned int len)
             return CL_BREAK;
         }
 
-        if (cli_hashset_contains(&iso->dir_blocks, block))
+        if (cli_hashset_contains(&iso->dir_blocks, block)) {
             continue;
+        }
 
-        if ((ret = cli_hashset_addkey(&iso->dir_blocks, block)) != CL_CLEAN)
+        if (CL_SUCCESS != (ret = cli_hashset_addkey(&iso->dir_blocks, block))) {
             return ret;
+        }
 
         dir = dir_orig = needblock(iso, block, 0);
-        if (!dir)
-            return CL_CLEAN;
+        if (!dir) {
+            return CL_SUCCESS;
+        }
 
         for (dirsz = MIN(iso->blocksz, len);;) {
             unsigned int entrysz = *dir, fileoff, filesz;
@@ -187,11 +192,8 @@ static int iso_parse_dir(iso9660_t *iso, unsigned int block, unsigned int len)
 
             cli_dbgmsg("iso_parse_dir: %s '%s': off %x - size %x - flags %x - unit size %x - gap size %x - volume %u\n", (dir[25] & 2) ? "Directory" : "File", iso->buf, fileoff, filesz, dir[25], dir[26], dir[27], cli_readint32(&dir[28]) & 0xffff);
             ret = cli_matchmeta(ctx, iso->buf, filesz, filesz, 0, 0, 0, NULL);
-            if (ret == CL_VIRUS) {
-                viruses_found = 1;
-                if (!SCAN_ALLMATCHES)
-                    break;
-                ret = CL_CLEAN;
+            if (ret != CL_SUCCESS) {
+                break;
             }
 
             if (dir[26] || dir[27])
@@ -201,59 +203,67 @@ static int iso_parse_dir(iso9660_t *iso, unsigned int block, unsigned int len)
                 if (dir[25] & 2) {
                     ret = iso_parse_dir(iso, fileoff, filesz);
                 } else {
-                    if (cli_checklimits("ISO9660", ctx, filesz, 0, 0) != CL_SUCCESS)
+                    if (CL_SUCCESS != cli_checklimits("ISO9660", ctx, filesz, 0, 0)) {
                         cli_dbgmsg("iso_parse_dir: Skipping overlimit file\n");
-                    else
+                    } else {
                         ret = iso_scan_file(iso, fileoff, filesz);
+                    }
                 }
-                if (ret == CL_VIRUS) {
-                    viruses_found = 1;
-                    if (!SCAN_ALLMATCHES)
-                        break;
-                    ret = CL_CLEAN;
+                if (ret != CL_SUCCESS) {
+                    break;
                 }
             }
             dirsz -= entrysz;
             dir += entrysz;
         }
 
-        fmap_unneed_ptr(*ctx->fmap, dir_orig, iso->blocksz);
+        fmap_unneed_ptr(ctx->fmap, dir_orig, iso->blocksz);
     }
-    if (viruses_found == 1)
-        return CL_VIRUS;
+
     return ret;
 }
 
-int cli_scaniso(cli_ctx *ctx, size_t offset)
+cl_error_t cli_scaniso(cli_ctx *ctx, size_t offset)
 {
     const uint8_t *privol, *next;
     iso9660_t iso;
     int i;
+    cl_error_t status   = CL_SUCCESS;
+    cl_error_t ret      = CL_SUCCESS;
+    uint32_t nextJoliet = 0;
 
-    if (offset < 32768)
-        return CL_CLEAN; /* Need 16 sectors at least 2048 bytes long */
+    if (offset < 32768) {
+        /* Need 16 sectors at least 2048 bytes long */
+        goto done;
+    }
 
-    privol = fmap_need_off(*ctx->fmap, offset, 2448 + 6);
-    if (!privol)
-        return CL_CLEAN;
+    privol = fmap_need_off(ctx->fmap, offset, 2448 + 6);
+    if (!privol) {
+        goto done;
+    }
 
     next = (uint8_t *)cli_memstr((char *)privol + 2049, 2448 + 6 - 2049, "CD001", 5);
-    if (!next)
-        return CL_CLEAN; /* Find next volume descriptor */
+    if (!next) {
+        /* Find next volume descriptor */
+        goto done;
+    }
 
     iso.sectsz = (next - privol) - 1;
-    if (iso.sectsz * 16 > offset)
-        return CL_CLEAN; /* Need room for 16 system sectors */
+    if (iso.sectsz * 16 > offset) {
+        /* Need room for 16 system sectors */
+        goto done;
+    }
 
     iso.blocksz = cli_readint32(privol + 128) & 0xffff;
     if (iso.blocksz != 512 && iso.blocksz != 1024 && iso.blocksz != 2048)
-        return CL_CLEAN; /* Likely not a cdrom image */
+        /* Likely not a cdrom image */
+        goto done;
 
     iso.base_offset = offset - iso.sectsz * 16;
     iso.joliet      = 0;
 
     for (i = 16; i < 32; i++) { /* scan for a joliet secondary volume descriptor */
-        next = fmap_need_off_once(*ctx->fmap, iso.base_offset + i * iso.sectsz, 2048);
+        next = fmap_need_off_once(ctx->fmap, iso.base_offset + i * iso.sectsz, 2048);
         if (!next)
             break; /* Out of disk */
         if (*next == 0xff || memcmp(next + 1, "CD001", 5))
@@ -283,60 +293,95 @@ int cli_scaniso(cli_ctx *ctx, size_t offset)
     /* TODO rr, el torito, udf ? */
 
     /* NOTE: freeing sector now. it is still safe to access as we don't alloc anymore */
-    fmap_unneed_off(*ctx->fmap, offset, 2448);
-    if (iso.joliet)
-        privol = next;
+    fmap_unneed_off(ctx->fmap, offset, 2448);
+    if (!iso.joliet) {
+        next = NULL;
+    }
 
-    cli_dbgmsg("in cli_scaniso\n");
-    if (cli_debug_flag) {
-        cli_dbgmsg("cli_scaniso: Raw sector size: %u\n", iso.sectsz);
-        cli_dbgmsg("cli_scaniso: Block size: %u\n", iso.blocksz);
+    nextJoliet = iso.joliet;
+    iso.joliet = 0;
 
-        cli_dbgmsg("cli_scaniso: Volume descriptor version: %u\n", privol[6]);
+    do {
+        cli_dbgmsg("in cli_scaniso\n");
+        if (cli_debug_flag) {
+            cli_dbgmsg("cli_scaniso: Raw sector size: %u\n", iso.sectsz);
+            cli_dbgmsg("cli_scaniso: Block size: %u\n", iso.blocksz);
+
+            cli_dbgmsg("cli_scaniso: Volume descriptor version: %u\n", privol[6]);
 
 #define ISOSTRING(src, len) iso_string(&iso, (src), (len))
-        cli_dbgmsg("cli_scaniso: System: %s\n", ISOSTRING(privol + 8, 32));
-        cli_dbgmsg("cli_scaniso: Volume: %s\n", ISOSTRING(privol + 40, 32));
+            cli_dbgmsg("cli_scaniso: System: %s\n", ISOSTRING(privol + 8, 32));
+            cli_dbgmsg("cli_scaniso: Volume: %s\n", ISOSTRING(privol + 40, 32));
 
-        cli_dbgmsg("cli_scaniso: Volume space size: 0x%x blocks\n", cli_readint32(&privol[80]));
-        cli_dbgmsg("cli_scaniso: Volume %u of %u\n", cli_readint32(privol + 124) & 0xffff, cli_readint32(privol + 120) & 0xffff);
+            cli_dbgmsg("cli_scaniso: Volume space size: 0x%x blocks\n", cli_readint32(&privol[80]));
+            cli_dbgmsg("cli_scaniso: Volume %u of %u\n", cli_readint32(privol + 124) & 0xffff, cli_readint32(privol + 120) & 0xffff);
 
-        cli_dbgmsg("cli_scaniso: Volume Set: %s\n", ISOSTRING(privol + 190, 128));
-        cli_dbgmsg("cli_scaniso: Publisher: %s\n", ISOSTRING(privol + 318, 128));
-        cli_dbgmsg("cli_scaniso: Data Preparer: %s\n", ISOSTRING(privol + 446, 128));
-        cli_dbgmsg("cli_scaniso: Application: %s\n", ISOSTRING(privol + 574, 128));
+            cli_dbgmsg("cli_scaniso: Volume Set: %s\n", ISOSTRING(privol + 190, 128));
+            cli_dbgmsg("cli_scaniso: Publisher: %s\n", ISOSTRING(privol + 318, 128));
+            cli_dbgmsg("cli_scaniso: Data Preparer: %s\n", ISOSTRING(privol + 446, 128));
+            cli_dbgmsg("cli_scaniso: Application: %s\n", ISOSTRING(privol + 574, 128));
 
 #define ISOTIME(s, n) cli_dbgmsg("cli_scaniso: "s                         \
                                  ": %c%c%c%c-%c%c-%c%c %c%c:%c%c:%c%c\n", \
                                  privol[n], privol[n + 1], privol[n + 2], privol[n + 3], privol[n + 4], privol[n + 5], privol[n + 6], privol[n + 7], privol[n + 8], privol[n + 9], privol[n + 10], privol[n + 11], privol[n + 12], privol[n + 13])
-        ISOTIME("Volume creation time", 813);
-        ISOTIME("Volume modification time", 830);
-        ISOTIME("Volume expiration time", 847);
-        ISOTIME("Volume effective time", 864);
+            ISOTIME("Volume creation time", 813);
+            ISOTIME("Volume modification time", 830);
+            ISOTIME("Volume expiration time", 847);
+            ISOTIME("Volume effective time", 864);
 
-        cli_dbgmsg("cli_scaniso: Path table size: 0x%x\n", cli_readint32(privol + 132) & 0xffff);
-        cli_dbgmsg("cli_scaniso: LSB Path Table: 0x%x\n", cli_readint32(privol + 140));
-        cli_dbgmsg("cli_scaniso: Opt LSB Path Table: 0x%x\n", cli_readint32(privol + 144));
-        cli_dbgmsg("cli_scaniso: MSB Path Table: 0x%x\n", cbswap32(cli_readint32(privol + 148)));
-        cli_dbgmsg("cli_scaniso: Opt MSB Path Table: 0x%x\n", cbswap32(cli_readint32(privol + 152)));
-        cli_dbgmsg("cli_scaniso: File Structure Version: %u\n", privol[881]);
+            cli_dbgmsg("cli_scaniso: Path table size: 0x%x\n", cli_readint32(privol + 132) & 0xffff);
+            cli_dbgmsg("cli_scaniso: LSB Path Table: 0x%x\n", cli_readint32(privol + 140));
+            cli_dbgmsg("cli_scaniso: Opt LSB Path Table: 0x%x\n", cli_readint32(privol + 144));
+            cli_dbgmsg("cli_scaniso: MSB Path Table: 0x%x\n", cbswap32(cli_readint32(privol + 148)));
+            cli_dbgmsg("cli_scaniso: Opt MSB Path Table: 0x%x\n", cbswap32(cli_readint32(privol + 152)));
+            cli_dbgmsg("cli_scaniso: File Structure Version: %u\n", privol[881]);
 
-        if (iso.joliet)
-            cli_dbgmsg("cli_scaniso: Joliet level %u\n", iso.joliet);
+            if (iso.joliet)
+                cli_dbgmsg("cli_scaniso: Joliet level %u\n", iso.joliet);
+        }
+
+        if (privol[156 + 26] || privol[156 + 27]) {
+            cli_dbgmsg("cli_scaniso: Interleaved root directory is not supported\n");
+            status = CL_SUCCESS;
+            goto done;
+        }
+
+        iso.ctx = ctx;
+        ret     = cli_hashset_init(&iso.dir_blocks, 1024, 80);
+        if (ret != CL_SUCCESS) {
+            status = ret;
+            goto done;
+        }
+
+        ret = iso_parse_dir(&iso, cli_readint32(privol + 156 + 2) + privol[156 + 1], cli_readint32(privol + 156 + 10));
+        cli_hashset_destroy(&iso.dir_blocks);
+        switch (ret) {
+            case CL_CLEAN:
+                /* fallthrough */
+            case CL_EFORMAT:
+                /* fallthrough */
+            case CL_EPARSE:
+                break;
+            default:
+                /*Recoverable errors are above, anything else we need to return.*/
+                status = ret;
+                goto done;
+        }
+        if (ret > status) {
+            status = ret;
+        }
+
+        privol     = next;
+        next       = NULL;
+        iso.joliet = nextJoliet;
+
+    } while (privol);
+
+done:
+
+    if (status == CL_BREAK) {
+        status = CL_SUCCESS;
     }
 
-    if (privol[156 + 26] || privol[156 + 27]) {
-        cli_dbgmsg("cli_scaniso: Interleaved root directory is not supported\n");
-        return CL_CLEAN;
-    }
-
-    iso.ctx = ctx;
-    i       = cli_hashset_init(&iso.dir_blocks, 1024, 80);
-    if (i != CL_SUCCESS)
-        return i;
-    i = iso_parse_dir(&iso, cli_readint32(privol + 156 + 2) + privol[156 + 1], cli_readint32(privol + 156 + 10));
-    cli_hashset_destroy(&iso.dir_blocks);
-    if (i == CL_BREAK)
-        return CL_CLEAN;
-    return i;
+    return status;
 }
